#
# Copyright (c) 2019-2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# This software product is a proprietary product of Nvidia Corporation and its affiliates
# (the "Company") and all right, title, and interest in and to the software
# product, including all associated intellectual property rights, are and
# shall remain exclusively with the Company.
#
# This software product is governed by the End User License Agreement
# provided with the software product.
#

from __future__ import print_function
import sys
import os
import subprocess
import re
from collections import namedtuple
import argparse
import shutil

MTUSB_DEVS_DIR = "/dev/mst/"
RETIMER_DEVS_DIR = "/dev/mst/retimer"

CMDS = ["add",
        "del",
        "status"]

# Append new retimers to this dictionary
RETIMERS_DIC = {
    "ArcusE": {
        "devid": 0xb200,
        "i2c_secondary_address": 0x50,
        "address_width": 2,
        "devid_address": 0xf0014
    }}


class Command(object):
    def __init__(self, cmd_str):
        self.cmd_str = cmd_str

    def execute(self):
        p = subprocess.Popen(self.cmd_str,
                             stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             universal_newlines=False,
                             shell=True)
        stdout, stderr = p.communicate()
        stat = p.wait()
        return stat, stdout, stderr


class I2cError(Exception):
    pass


class Retimer(object):

    @classmethod
    def discover_external_chips(cls):
        cable_devices, mtusb_devices = [], []
        for mst_device in os.listdir(MTUSB_DEVS_DIR):
            if 'cable' in mst_device:
                cable_devices.append(mst_device)
            elif 'mtusb' in mst_device:
                mtusb_devices.append(MTUSB_DEVS_DIR + mst_device)

        retimers_in_cables_devs = cls._discover_chips_in_cables(cable_devices)
        retimers_in_evb_devs = cls._discover_chip_on_EVB(mtusb_devices)
        return retimers_in_cables_devs + retimers_in_evb_devs

    @classmethod
    def _discover_chips_in_cables(cls, cable_devices):
        """
        Discover retimers in a module via MCU (FW Gateway)
        The function returns a list with the name of discovered retimer devices
        """
        linkx_devices = []
        # TODO: Support retimers on cable discovery
        return linkx_devices

    @classmethod
    def _discover_chip_on_EVB(cls, mtusb_devices):
        """
        Check if the device connected to MTUSB is a board with a retimer chip
        The function returns a list with the name of the discovered retimers devices
        """
        RetimerDevice = namedtuple("RetimerDevice", 'devid i2c_secondary_address address_width devid_address')
        retimers_to_scan = []
        for dev in RETIMERS_DIC:
            retimers_to_scan.append(RetimerDevice(RETIMERS_DIC[dev]["devid"],
                                                  RETIMERS_DIC[dev]["i2c_secondary_address"],
                                                  RETIMERS_DIC[dev]["address_width"],
                                                  RETIMERS_DIC[dev]["devid_address"]))

        linkx_devices = []
        for mtusb_device in mtusb_devices:
            # Create a list of known i2c secondary (Linkx) devices
            LinkxDevice = namedtuple(
                'LinkxDevice', 'devid i2c_secondary_address address_width devid_address')

            arcuse_rev0 = LinkxDevice(
                devid=0xb200, i2c_secondary_address=0x50, address_width=2, devid_address=0xf0014)

            i2c_secondarys = self._i2c_scan(mtusb_device, retimers_to_scan)

            if len(i2c_secondarys) == 1:
                retimer_dev_name = "{}_rt{}".format(mtusb_device.replace(
                    "/dev/mst/", ""), int(i2c_secondarys[0].devid, base=16))
                retimer_devs.append(retimer_device_name)
            elif len(i2c_secondarys) > 1:
                raise Exception(
                    'i2c primary detected more than one i2c secondaries devices')

        return linkx_devices

    def _i2c_cmd_read(self, mtusb_device, i2c_secondary_address, address_width, address, size):
        cmd = "i2c -a {0} -x {1} {2} read {3:#x} {4:#x}".format(
            address_width, size, mtusb_device, i2c_secondary_address, address)

        rc, out, err = Command(cmd).execute()

        if rc == 0:
            out = out.strip()
            result = []
            for index in range(len(out) - 2, -1, -2):
                byte_value = int(out[index:index + 2], 16)
                result.append(byte_value)
            return result
        else:
            raise I2cError("Failed to read from i2c secondary address {0} ({1})".format(
                i2c_secondary_address, err))

    def _i2c_scan(self, mtusb_device, devs_to_scan):
        result = []
        for device in devs_to_scan:
            try:
                devid = self._i2c_cmd_read(
                    mtusb_device, device.i2c_secondary_address, device.address_width, device.devid_address, 4)[0]
                if devid == device.devid:
                    result.append(device)
            except I2cError:
                pass
        return result


if __name__ == '__main__':
    try:
        parser = argparse.ArgumentParser(prog="mst_retimer",
                                         description="Retimers discovery script")
        parser.add_argument("command", choices=CMDS)
        args = parser.parse_args()

        if args.command == "del":
            if os.path.exists(RETIMER_DEVS_DIR):
                shutil.rmtree(RETIMER_DEVS_DIR,
                              ignore_errors=False, onerror=None)

        elif args.command == "add":
            retimers_ext_devs = Retimer.discover_external_chips()
            chipset_cnt = 0
            for dev in retimers_ext_devs:
                dev_file = os.path.join(RETIMER_DEVS_DIR, file)
                if not os.path.exists(dev_file):
                    open(dev_file, 'w').close()
                    chipset_cnt += 1

            if chipset_cnt:
                print("-I- Added {} retimer device{} ..".format(chipset_cnt,
                                                                "s" if chipset_cnt > 1 else ""))
            else:
                print("-I- No Retimer devices found")

        elif args.command == "status":
            if os.path.exists(RETIMER_DEVS_DIR):
                print("Retimer devices:")
                print("----------------")
                for file in os.listdir(RETIMER_DEVS_DIR):
                    print(os.path.join(RETIMER_DEVS_DIR, file))

    except Exception as err:
        print(err, file=sys.stderr)
        sys.exit(1)
